@@ -9,20 +9,32 @@ class ProtocolError(Exception):
 class CredentialsError(Exception):
     pass
 
+class ConnectionError(Exception):
+    pass
+
 class SubscriptionIterator:
     def __init__(self):
         self.queue = asyncio.Queue()
     
     # TODO: should we enqueue raw messages instead?
     def enqueue_payload(self, payload):
-        self.queue.put_nowait(payload)
+        queue_item = ('payload', payload)
+        self.queue.put_nowait(queue_item)
     
+    def close_iterator(self):
+        self.queue.put_nowait(('stop',))
+
     def __aiter__(self):
         return self
     
     async def __anext__(self):
-        payload = await self.queue.get()
-        return payload
+        queue_item = await self.queue.get()
+        if queue_item[0] == 'payload':
+            return queue_item[1]
+        elif queue_item[0] == 'stop':
+            raise StopAsyncIteration
+        else:
+            raise NotImplementedError()
 
 class StreamSeaConnection:
     def __init__(self, *, remote_host, remote_port=None, secure=False, client_id, client_secret, loop=None):
@@ -87,9 +99,15 @@ class StreamSeaConnection:
     
     async def authenticated_listen(self):
         # Receive messages forever
-        while True:
-            r = await self.recv()
-            await self.on_authenticated_message(r)
+        try:
+            while True:
+                r = await self.recv()
+                await self.on_authenticated_message(r)
+        except:
+            for k, replies_map_entry in self.replies_map.items():
+                if not replies_map_entry['first_reply_future'].done():
+                    replies_map_entry['first_reply_future'].set_exception(ConnectionError())
+                replies_map_entry['other_replies_iterator'].close_iterator()
 
     async def recv(self):
         ws = await self.ensure_ws
@@ -108,10 +126,8 @@ class StreamSeaConnection:
         other_replies_iterator = SubscriptionIterator()
 
         self.replies_map[msg_id] = {
-            "type": "multi_reply",
             "first_reply_future": first_reply_future,
             "other_replies_iterator": other_replies_iterator,
-            "received_reply": False,
         }
         await self.send(subscription_request_msg_json)
         
@@ -133,14 +149,10 @@ class StreamSeaConnection:
                 raise ProtocolError('Server sent a response but the message id could not be resolved to a request')
 
             replies_map_entry = self.replies_map[msg_id]
-            if replies_map_entry['type'] == 'multi_reply':
-                if replies_map_entry['received_reply']:
-                    replies_map_entry['other_replies_iterator'].enqueue_payload(message_json['payload'])
-                else:
-                    replies_map_entry['received_reply'] = True
-                    replies_map_entry['first_reply_future'].set_result(message_json)
+            if replies_map_entry['first_reply_future'].done():
+                replies_map_entry['other_replies_iterator'].enqueue_payload(message_json['payload'])
             else:
-                pass # TODO
+                replies_map_entry['first_reply_future'].set_result(message_json)
 
         except json.decoder.JSONDecodeError:
             raise ProtocolError('Could not decode JSON for message')
